#include <string.h>
#include <tx_api.h>
#include "math.h"
#include "ampire_graphics.h"
#include "ampire_utils.h"
#include "ampire_cfg.h"

static unsigned short tempImage[SCREEN_WIDTH * SCREEN_HEIGHT ];

/******************
 * Bitmap stuffs
 * ****************/
#define COLOR_PALATTE_SIZE	256
 
struct __attribute__ ((__packed__)) bitmap_header
{
	unsigned short bfType; // should decimal value 19778
	unsigned long bfSize; // size of the file in bytes
	char reserved[4];
	unsigned long bfOffBits; // specifies the offset from the beginning of the file to the bitmap data
}; //14 bytes in length


struct __attribute__ ((__packed__)) bitmap_info_header
{
	unsigned long biSize; // size of this structure, should be 40
	unsigned long biWidth; // width of image
	unsigned long biHeight; // height of image
	unsigned short biPlanes; // number of planes zooming through the air!
	unsigned short biBitCount; // number of bits per pixel, we expect 8
	unsigned long biCompression; // type of compression, normal is zero
	unsigned long biSizeImage; // size of image, if no compression, this can be zero
	unsigned long biXPelsPerMeter;
	unsigned long biyPelsPerMeter;
	unsigned long biClrUsed; // number of colors used
	unsigned long biClrImportant;
}; // 40 bytes in length

static unsigned long colorPalette[COLOR_PALATTE_SIZE];
static unsigned short convertToPalete(unsigned long data)
{
	
	unsigned short retval = 0, tval;
	char *mdata = (char *) &data;
	
	tval = (unsigned short) ((mdata[1] >> 3) << 11);  //red
	retval |= tval;

	tval = (unsigned short) ((mdata[2] >> 2) << 5); //green
	retval |= tval;

	tval = (unsigned short) (mdata[3] >> 3); //blue
	retval |= tval;

	return retval;
}


/*
If bitmap_info_header.biBitCount == 16 (2^16 colors):
	If bInfoHeader.biCompression == bi_rgb (0)
		the palette field does not contain any entries. 
		each word in the bitmap array represents a single pixel. 
		the relative intensities of RGB are represented with 5 bits for each color component.
		the value for blue is in the least significant 5 bits, followed by 5 bits each for 
		green and red, respectively. the most significant bit is not used.
	if bInfoHeader.biCompression == bi_bitfields(3), 
		the palette field contains three dword color masks that specify the red, green, and 
		blue components, respectively, of each pixel. each word in the bitmap array represents
		a single pixel.
		The 16bpp color masks can be:
			5-5-5, where B is 0x001f, G is 0x03e0, R is 0x7c00; 
			5-6-5, where B is 0x001f, G is 0x07e0, R is 0xf800.
 * */

void amp_draw_bitmap(unsigned char *bitmap)
{
	struct bitmap_header bHeader;
	struct bitmap_info_header bInfoHeader;
	unsigned char *writePos;
	unsigned short *pshort, *pBuffer;
	unsigned long *pPalette, *pPalette_bis;
	unsigned long startpos;
	int x, len, y;

	getMutex();

	memcpy(&bHeader, bitmap, sizeof(bHeader));

	bHeader.bfType = swap_short(bHeader.bfType);
	bHeader.bfSize = swap_long(bHeader.bfSize);
	bHeader.bfOffBits = swap_long(bHeader.bfOffBits);

	memcpy(&bInfoHeader, bitmap + sizeof(bHeader), sizeof(bInfoHeader));

	bInfoHeader.biSize 			= swap_long(bInfoHeader.biSize);
	bInfoHeader.biWidth 		= swap_long(bInfoHeader.biWidth);
	bInfoHeader.biHeight 		= swap_long(bInfoHeader.biHeight);
	bInfoHeader.biBitCount 		= swap_short(bInfoHeader.biBitCount);
	bInfoHeader.biSizeImage 	= swap_long(bInfoHeader.biSizeImage);
	bInfoHeader.biClrUsed 		= swap_long(bInfoHeader.biClrUsed);
	bInfoHeader.biClrImportant 	= swap_long(bInfoHeader.biClrImportant);
	bInfoHeader.biCompression  	= swap_long(bInfoHeader.biCompression);

	if(bInfoHeader.biCompression == 1 || bInfoHeader.biCompression == 2 || bInfoHeader.biBitCount == 24)
		goto _goOut;//Compression or format not supported
	
	startpos = sizeof(bHeader) + sizeof(bInfoHeader);

	if(  bInfoHeader.biBitCount == 16)
	{//16 BPP
		len = (bHeader.bfSize - (54 + 12))/2;
		pshort =(unsigned short*) (bitmap + bHeader.bfOffBits);		
		y=3;//size of "palette field"		
	}
	else if( bInfoHeader.biBitCount == 8)
	{//8 BPP
	    len = bHeader.bfSize - (54 + 1024);			
	    writePos = bitmap + bHeader.bfOffBits;		
		y = COLOR_PALATTE_SIZE-1;//size of "palette field"		
	}
	
	pBuffer = tempImage + sizeof(tempImage)/2 - 1;
	pPalette = colorPalette;
	pPalette_bis = (unsigned long*)(bitmap + startpos);
	
	//load color palatte if 8bit or get RGB (5-6-5)mask's color (0xf8, 0x7e0,0x1f) if 16bit.		
	for(x = 0; x < y; x++){	
		memcpy(pPalette, pPalette_bis++, 4);
		*pPalette = swap_long(*pPalette);
		pPalette++;
	}
	
	if( (bInfoHeader.biWidth == SCREEN_WIDTH) && (bInfoHeader.biHeight == SCREEN_HEIGHT) )
	{
		if( bInfoHeader.biBitCount == 16)
		{
			for(x = 0; x < len; x++)
				*pBuffer-- =	swap_short( *pshort++ );
		}
		else if( bInfoHeader.biBitCount == 8)
		{				
			for(x = 0; x < len; x++)
				*pBuffer-- = convertToPalete(colorPalette[(int) *writePos++]);
		}	
	}
	else if( (bInfoHeader.biWidth == SCREEN_HEIGHT) && (bInfoHeader.biHeight == SCREEN_WIDTH) )
	{
		if( bInfoHeader.biBitCount == 16)
		{
			for(x = (SCREEN_WIDTH - 1); x >= 0; x--)
				for(y = 0; y < SCREEN_HEIGHT; y++)
					tempImage[(x + (y * SCREEN_WIDTH))] = swap_short( *pshort++ );
		}
		else if( bInfoHeader.biBitCount == 8)
		{		
			for(x = (SCREEN_WIDTH - 1); x >= 0; x--)
				for(y = 0; y < SCREEN_HEIGHT; y++)
					tempImage[(x + (y * SCREEN_WIDTH))] = convertToPalete(colorPalette[(int) *writePos++]);
		}						
	}
	else
		goto _goOut;

	amp_write_data(0, SCREEN_WIDTH, 0, SCREEN_HEIGHT, tempImage);

_goOut:
	releaseMutex();
    return;
}


/*********************
 *
 *  Draw utilities 
 * 
 **********************/


void amp_fillscreen(unsigned short color)
{
	int x;

	getMutex();
	
	for(x = 0; x < (SCREEN_WIDTH * SCREEN_HEIGHT); x++)
		tempImage[x] = color;

	amp_write_data(0, SCREEN_WIDTH, 0, SCREEN_HEIGHT, tempImage);

	releaseMutex();

	return;
}

void amp_rectangle(point_t *a, point_t *b, unsigned short color, int solid)
{
	int z,l,h ;
	long lh;
	
	if(a->x >= b->x || a->y >= b->y)
		return;

	l = b->x - a->x + 1;
	h = b->y - a->y + 1;
	lh= l* h;
	
	getMutex();

	if(solid)
	{
		for(z = 0 ; z <= lh ; z++)
			tempImage[z] = color;
		amp_write_data(a->x, b->x, a->y, b->y, tempImage);
	}
	else{
		for(z = 0 ; z < (l+h) ; z++)
			tempImage[z] = color;
		
		amp_write_data(a->x, b->x, a->y, a->y, tempImage);
		amp_write_data(a->x, b->x, b->y, b->y, tempImage);
		amp_write_data(a->x, a->x, a->y, b->y, tempImage);
		amp_write_data(b->x, b->x, a->y, b->y, tempImage);
	}
	

	releaseMutex();

	return;
}




void amp_draw_pixel(point_t *p, unsigned short color)
{
	getMutex();
	amp_write_data(p->x, p->x, p->y, p->y, &color);
	releaseMutex();
	return;
}


void amp_amp_draw_image(unsigned int left, unsigned int right, unsigned int top, unsigned int bottom, unsigned short *colors)
{
	getMutex();
	amp_write_data(left, right, top, bottom, colors);
	releaseMutex();
	return;
}

typedef struct {
	float x,y;
} fpoint_t;


void amp_line(point_t *a, point_t *b, unsigned short color)
{
	fpoint_t v;//float point   
	point_t p;
    float aux;
    int l; 
	
	//transfer vector to the origin
	v.x =  b->x - a->x;
	v.y =  b->y - a->y;
	
	//calculate versor
	aux	= sqrt( (v.x*v.x) +  (v.y*v.y) );
	v.x /= aux;
	v.y /= aux;
	getMutex();
	
	l=0;
	do
	{
		p.x = (int)(((l/10) * v.x) + a->x + .5);	
		p.y = (int)(((l/10) * v.y) + a->y + .5); 			
		setxypos(p.x, p.x, p.y, p.y);
		write_cmd(GCTRL_RAMWR);
		
		write_ram( color );			
		l++;
	}while((v.x>0 && p.x<b->x) || (v.x<0 && p.x>b->x) || (v.y==1 && p.y<b->y) || (v.y==-1 && p.y>b->y));
	
	write_cmd(GCTRL_RAMRD);	
	
	releaseMutex();

    return;
}

void amp_circle(point_t *p, int r,  unsigned short color, int solid )
{
	int i,j,ca, y_up, y_down;
	float co;
	
	getMutex();
	for(i = p->x-r ; i<=p->x+r ; i++ )
	{
		ca = i-p->x;
		co = sqrt(r*r - ca*ca);
		
		//rounding
		y_up = (int)(p->y + co + .5);
		y_down = (int)(p->y - co + .5);
						
		setxypos(i, i, y_up, y_up);
		write_cmd(GCTRL_RAMWR);
		write_ram( color );					
		
		setxypos(i, i,y_down, y_down);
		write_cmd(GCTRL_RAMWR);
		write_ram( color );				
							
		if(solid){
			setxypos(i, i, y_down+1,y_up-1);
			write_cmd(GCTRL_RAMWR);			
			for(j=y_down+1; j<y_up; j++){
				write_ram( color );
			}
		}		
	}

	write_cmd(GCTRL_RAMRD);	
	releaseMutex();
	
}


/****************************
 * 
 * 	DISPLAY TEST FUNCTIONS
 *  
 ****************************/

void amp_color_test(unsigned short * image)
{
	int r=0,g=0,b=0;
	unsigned int y,i=0;	
	
	if(image==NULL)
		image = tempImage;
		
	//From R+G+B to R
	for(r=31,b=31,g=63; g>0; b--,g-=2){
		for(y=0; y< SCREEN_HEIGHT ; y++)
			image[y*SCREEN_WIDTH+i] = (r<<11) | (g<<5) | (b<<0);	
		i++;				
	}	

	//From R to R+G
	for(r = 31,g=0,b=0; g<64; g+=2){
		for(y=0; y< SCREEN_HEIGHT ;y++ )
			image[y*SCREEN_WIDTH+i] = (r<<11) | (g<<5) | (b<<0);	
		i++;		
	}
	
	//From R+G to G	 
	for(g=63; r>=0; --r){
		for(y=0; y< SCREEN_HEIGHT ;y++ )
			image[y*SCREEN_WIDTH+i] = (r<<11) | (g<<5) | (b<<0);	
		i++;		
	}
	
	//From G to G+B
	for(r=0; b<32; ++b){
		for(y=0; y< SCREEN_HEIGHT ;y++ )
			image[y*SCREEN_WIDTH+i] = (r<<11) | (g<<5) | (b<<0);	
		i++;				
	}
	
	//From G+B to B
	for(b=31; g>=0; --g){
		for(y=0; y< SCREEN_HEIGHT ; y++)
			image[y*SCREEN_WIDTH+i] = (r<<11) | (g<<5) | (b<<0);	
		i++;				
	}	
	
	//From B to B+R
	for(g=0; r<32; ++r){
		for(y=0; y< SCREEN_HEIGHT ; y++)
			image[y*SCREEN_WIDTH+i] = (r<<11) | (g<<5) | (b<<0);	
		i++;				
	}		
	
	//From B+R to R
	for(r=31; b>=0 ; --b){
		for(y=0; y< SCREEN_HEIGHT ; y++)
			image[y*SCREEN_WIDTH+i] = (r<<11) | (g<<5) | (b<<0);	
		i++;				
	}
	
	//From R to R+(G/2)
	for(b=0; g<32 ; g++){
		for(y=0; y< SCREEN_HEIGHT ; y++)
			image[y*SCREEN_WIDTH+i] = (r<<11) | (g<<5) | (b<<0);	
		i++;				
	}
		
	//From R to R+G+B
	for(; g<64 && i<SCREEN_WIDTH; g++,b++){
		for(y=0; y< SCREEN_HEIGHT ; y++)
			image[y*SCREEN_WIDTH+i] = (r<<11) | (g<<5) | (b<<0);	
		i++;				
	}	
	
	getMutex();
	amp_write_data(0, SCREEN_WIDTH, 0, SCREEN_HEIGHT, image);
	releaseMutex();

	return;
}



int amp_rainbow_fps_test(int frames)
{

	int iterations = 0;
	unsigned long start_time, end_time;
	static unsigned short buffer[(SCREEN_HEIGHT * SCREEN_WIDTH) + EXTRA_SPACE];

	if(frames > EXTRA_SPACE)
		return 0;

	getMutex();

	amp_color_test(buffer);

	memcpy(&buffer[SCREEN_HEIGHT * SCREEN_WIDTH], buffer, EXTRA_SPACE);

	start_time = tx_time_get();

	while(iterations < frames)
	{
		amp_write_data(0, SCREEN_WIDTH, 0, SCREEN_HEIGHT, (&buffer[iterations]));
		iterations++;
	}

	end_time = tx_time_get();

	releaseMutex();

	return (end_time - start_time);
}

#include "fonts.h"


void amp_write_at(unsigned int x, unsigned int y, char * t, unsigned short fontcolor, unsigned short backcolor, int size)
{
	unsigned int left, right, top, bottom;
	int row,col, steps,aux; 
	const unsigned char* tOffset;
	const unsigned char* charset;
	
	char FONT_W, FONT_H;	

	FONT_W = (size>>4)&0x0f;
	FONT_H = size&0x0f;
	bottom = y + FONT_H;		
	top = y;	
	left  = x;
	right = left + FONT_W - 1;
	switch(size){
		case FONT_SIZE_5x7:
			size = 5;
			charset = font5x7;
			aux=1;
			break;
		case FONT_SIZE_8x8:
			size = 8;	
			charset = font8x8; 
			aux=1;			
			break;		
		case FONT_SIZE_8x15:
			size = 8;		
			charset = font8x15;		
			aux=2;			
			break;		
	}
	
	
	while(*t)
	{	
 		if(left >= (SCREEN_WIDTH - FONT_W - FONT_MARGIN) ){
 			left = x;
 			right = left + FONT_W - 1;
 			top += FONT_H;
 			bottom += FONT_H; 
 		}

		setxypos(left, right , top  , bottom );
		
		tOffset = charset + *t * (size*aux); 
		
		steps  = 0;
		
		//Write Enable  for AMP Display	
		write_cmd(GCTRL_RAMWR);
		//write one character
		while(steps<aux){
			for(row=0;row<8; row++){
				for(col=0; col<size; col++ ){	
					if(*(tOffset+ col + 8*steps ) & (1<<row))
						write_ram( fontcolor );
					else{
						if(backcolor==fontcolor)
							read_data();
						else					
							write_ram( backcolor ); 
					}
				}
			}		
			steps++;								
		}				
		//Write Disable  for AMP Display
		write_cmd(GCTRL_RAMRD);
		
		left  += FONT_W;
		right = left + FONT_W - 1;		
		t++;		
	}
	

}





